#include <QWidget>
#include <QVBoxLayout>
#include <QLineEdit>
#include <QPushButton>
#include <QTextEdit>
#include <QLabel>
#include <QString>
#include <QThread>
#include <QCheckBox>
#include <QHBoxLayout>
#include <cmath>
#include <QSpinBox>
#include <QApplication>

using namespace std;
using namespace std::chrono;
void merge(vector<int>& arr, int left, 
           int mid, int right)
{
    int n1 = mid - left + 1;
    int n2 = right - mid;
    vector<int> L(n1), R(n2);
    
    for (int i = 0; i < n1; i++)
        L[i] = arr[left + i];
    for (int j = 0; j < n2; j++)
        R[j] = arr[mid + 1 + j];
    
    int i = 0, j = 0;
    int k = left;
    while (i < n1 && j < n2) {
        if (L[i] <= R[j]) {
            arr[k] = L[i];
            i++;
        }
        else {
            arr[k] = R[j];
            j++;
        }
        k++;
    }
    while (i < n1) {
        arr[k] = L[i];
        i++;
        k++;
    }
    while (j < n2) {
        arr[k] = R[j];
        j++;
        k++;
    }
}
void mergeSort(vector<int>& arr, int left, int right)
{
    if (left >= right)
        return;
    
    int mid = left + (right - left) / 2;
    mergeSort(arr, left, mid);
    mergeSort(arr, mid + 1, right);
    merge(arr, left, mid, right);
}
class SortWorker : public QObject {
    Q_OBJECT
public:
    SortWorker(vector<int>& arr, unsigned int n, bool individualTask) 
        : arr(arr), n(n), individualTask(individualTask), shouldStop(false) {}
    
    void stop() {
        shouldStop = true;
    }
    
public slots:
    void performSort() {
        auto start_time = high_resolution_clock::now();
        mergeSort(arr,0,arr.size()-1);
        auto duration = duration_cast<microseconds>(high_resolution_clock::now() - start_time);
        emit progressUpdated(100, n, n);
        emit sortingFinished(duration.count(), is_sorted(arr.data(), n, [](int a, int b) { return a < b; }));
    }
    
signals:
    void progressUpdated(int progress, unsigned int current, unsigned int total);
    void sortingFinished(int duration, bool sorted);
    void sortingCancelled();
    
private:
    vector<int>& arr;
    unsigned int n;
    bool individualTask;
    std::atomic<bool> shouldStop;
    template<typename Compare>
    bool is_sorted(int* arr, unsigned int n, Compare comp) {
        for (unsigned int i = 0; i < n - 1; i++) {
            if (comp(arr[i+1], arr[i])) return false;
        }
        return true;
    }
};
class SortWorkerWorse : public QObject {
    Q_OBJECT
public:
    SortWorkerWorse(vector<int>& arr, unsigned int n, bool individualTask) 
        : arr(arr), individualTask(individualTask), shouldStop(false) {}
    
    void stop() {
        shouldStop = true;
    }
    
public slots:
    void performSort() {
        unsigned int firstmin;
        unsigned int lastmin;
        for (firstmin=0;arr[firstmin]>=0;firstmin++){}
        for (lastmin=arr.size()-1;arr[lastmin]>=0;lastmin--){}
        auto start_time = high_resolution_clock::now();
        mergeSort(arr,firstmin,lastmin);
        auto duration = duration_cast<microseconds>(high_resolution_clock::now() - start_time);
        emit progressUpdated(100, arr.size(), arr.size());
        emit sortingFinished(duration.count(),1);
    }
    
signals:
    void progressUpdated(int progress, unsigned int current, unsigned int total);
    void sortingFinished(int duration, bool sorted);
    void sortingCancelled();
    
private:
    vector<int>& arr;
    bool individualTask;
    std::atomic<bool> shouldStop;
};

class SortingApp : public QWidget {
    Q_OBJECT
public:
    SortingApp(QWidget *parent = nullptr) : QWidget(parent), worker(nullptr), wworker(nullptr), isSorting(false) {
        setupUI();
    }

    ~SortingApp() {
        if (workerThread.isRunning()) {
            workerThread.quit();
            workerThread.wait();
        }
    }

private:
    QSpinBox *inputField;
    QPushButton *sortButton;
    QTextEdit *outputField;
    QCheckBox *showArrayCheckBox;
    vector<int> arr;
    QThread workerThread;
    SortWorker *worker;
    SortWorkerWorse *wworker;
    QCheckBox *individualTaskCheckBox;
    bool isSorting;

    void setupUI() {
        QVBoxLayout *mainLayout = new QVBoxLayout(this);

        QLabel *label = new QLabel("Скільки чисел сортувати?", this);
        mainLayout->addWidget(label);

        inputField = new QSpinBox(this);
        inputField->setRange(1, 10000000);
        inputField->setValue(10);  // Встановлюємо початкове значення
        mainLayout->addWidget(inputField);

        sortButton = new QPushButton("Сортувати", this);
        connect(sortButton, &QPushButton::clicked, this, &SortingApp::onSortButtonClicked);
        mainLayout->addWidget(sortButton);

        QHBoxLayout *checkBoxLayout = new QHBoxLayout();
        
        showArrayCheckBox = new QCheckBox("Показати масив до та після сортування", this);
        checkBoxLayout->addWidget(showArrayCheckBox);

        individualTaskCheckBox = new QCheckBox("Індивідуальне завдання", this);
        checkBoxLayout->addWidget(individualTaskCheckBox);

        mainLayout->addLayout(checkBoxLayout);

        outputField = new QTextEdit(this);
        outputField->setReadOnly(true);
        mainLayout->addWidget(outputField);

        setLayout(mainLayout);
        setWindowTitle("Сортування масиву");
        resize(500, 500);
    }

    void sortArray() {
        arr.clear();
        arr.resize(inputField->value());
        for (unsigned int i = 0; i < inputField->value(); i++) arr[i] = rand() % 20000001 - 10000000;
        outputField->clear();
        if (showArrayCheckBox->isChecked()) {
            outputField->append("Масив до сортування:");
            outputField->append(arrayToString(arr));
        }
        if (individualTaskCheckBox->isChecked()){
            wworker = new SortWorkerWorse(arr, inputField->value(), individualTaskCheckBox->isChecked());
            wworker->moveToThread(&workerThread);
            outputField->append("==Режим індивідуального завдання #9==\n");
            connect(this, &SortingApp::startSort, wworker, &SortWorkerWorse::performSort);
            connect(wworker, &SortWorkerWorse::progressUpdated, this, &SortingApp::updateProgress);
            connect(wworker, &SortWorkerWorse::sortingFinished, this, &SortingApp::finishSorting);
            connect(&workerThread, &QThread::finished, wworker, &QObject::deleteLater);
        }else{
            worker = new SortWorker(arr, inputField->value(), individualTaskCheckBox->isChecked());
            worker->moveToThread(&workerThread);
            connect(this, &SortingApp::startSort, worker, &SortWorker::performSort);
            connect(worker, &SortWorker::progressUpdated, this, &SortingApp::updateProgress);
            connect(worker, &SortWorker::sortingFinished, this, &SortingApp::finishSorting);
            connect(&workerThread, &QThread::finished, worker, &QObject::deleteLater);}
        workerThread.start();
        emit startSort();
    }
    
    QString arrayToString(const vector<int>& arr) {
        QString result;
        for (size_t i = 0; i < arr.size(); ++i) {
            result += QString::number(static_cast<int>(arr[i]));
            if (i < arr.size() - 1) result += ", ";
        }
        return result;
    }
private slots:
    void onSortButtonClicked() {
        if (!isSorting) {
        isSorting = true;
        sortButton->setText("Скасувати");
            sortArray();
        } else resetSortingState();
    }

    void updateProgress(int progress, unsigned int current, unsigned int total) {
        outputField->append(QString("Відсортовано %1% (%2 з %3 елементів)").arg(progress).arg(current).arg(total));
    }

    void finishSorting(int duration, bool sorted) {
        outputField->append(QString("\nЧас сортування: %1 мікросекунд").arg(duration));
        outputField->append(sorted ? "Масив відсортовано правильно." : "Помилка: масив не відсортовано правильно.");
        
        if (showArrayCheckBox->isChecked()) {
            outputField->append("\nМасив після сортування:");
            outputField->append(arrayToString(arr));
            arr.clear();
        }
        
        resetSortingState();
    }
    void resetSortingState() {
        isSorting = false;
        sortButton->setText("Сортувати");
        if(worker){
            worker->disconnect();worker->deleteLater();}
        if(wworker){wworker->deleteLater();
            wworker->disconnect();}
        worker=nullptr;
        wworker=nullptr;
    }

signals:
    void startSort();
};

int main(int argc, char *argv[]) {
    QApplication app(argc, argv);
    SortingApp sortingApp;
    sortingApp.show();
    return app.exec();
}
#include "main.moc"
